package com.google.code.joliratools.logdb.parser;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;

import com.google.code.joliratools.logdb.parser.Record.Type;

final class DelimitedParser implements Parser {
    private static class DelimitedRecord implements Record {
        private final String level;
        private final String[] variables;
        private final long timestamp;
        private final String clazz;
        private final String method;
        private final int lineNo;
        private final StringBuilder raw;
        private final Type type;

        DelimitedRecord(final Type type, final int lineNo, final String raw, final String[] variables,
                final long timestamp, final String clazz, final String method, final String level) {
            this.type = type;
            this.lineNo = lineNo;
            this.raw = new StringBuilder();
            this.variables = variables;
            this.timestamp = timestamp;
            this.clazz = clazz;
            this.method = method;
            this.level = level;

            this.raw.append(raw);
        }

        public void append(final String line) {
            raw.append(NEWLINE);
            raw.append(line);

            if (variables == null || variables.length < 1) {
                return;
            }

            final StringBuilder buf = new StringBuilder();

            buf.append(variables[variables.length - 1]);
            buf.append(NEWLINE);
            buf.append(line);

            variables[variables.length - 1] = buf.toString();
        }

        @Override
        public String getClazz() {
            return clazz;
        }

        @Override
        public String getId() {
            return "" + clazz + '#' + method + '[' + timestamp + ']';
        }

        @Override
        public String getLevel() {
            return level;
        }

        @Override
        public int getLineNumber() {
            return lineNo;
        }

        @Override
        public String getMethod() {
            return method;
        }

        @Override
        public String getRaw() {
            return raw.toString();
        }

        @Override
        public long getTimestamp() {
            return timestamp;
        }

        @Override
        public Type getType() {
            return type;
        }

        @Override
        public String[] getVariables() {
            return variables;
        }
    }

    final static String NEWLINE = System.getProperty("line.separator");

    private final BufferedReader reader;
    private DelimitedRecord active = null;
    private int lineNo = 0;
    private final String fieldDelimiter;
    private final int timeStampPos;
    private final int classPos;
    private final int methodPos;
    private final String timeStampFormat;
    private final int maxPos;
    private final int levelPos;

    DelimitedParser(final InputStream in, final String fieldDelimiter, final String timeStampFormat,
            final int timeStampPos, final int classPos, final int methodPos, final int levelPos, final int maxPos) {
        this.fieldDelimiter = fieldDelimiter;
        this.timeStampFormat = timeStampFormat;
        this.timeStampPos = timeStampPos;
        this.classPos = classPos;
        this.methodPos = methodPos;
        this.levelPos = levelPos;
        this.maxPos = maxPos;

        reader = new BufferedReader(new InputStreamReader(in));
    }

    private String getField(final String[] fields, final int pos) {
        if (pos < 0) {
            return null;
        }

        return fields[pos];
    }

    private String[] getFields(final String line) {
        final Collection<String> vars = new ArrayList<String>();
        int start = 0;

        for (;;) {
            final int end = line.indexOf(fieldDelimiter, start);

            if (end == -1) {
                break;
            }

            final String field = start == end ? null : line.substring(start, end);

            vars.add(field);

            start = end + 1;
        }

        final String field = line.substring(start);

        vars.add(field);

        final int size = vars.size();

        return vars.toArray(new String[size]);
    }

    @Override
    public synchronized Record getNext() throws ParseException, IOException {
        for (;;) {
            lineNo++;
            final String line = reader.readLine();
            final DelimitedRecord current = active;

            if (line == null) {
                active = null;
                return current;
            }

            final DelimitedRecord _new = parse(line);

            if (_new != null) {
                active = _new;

                if (current != null) {
                    return current;
                }

                continue;
            }

            if (current == null) {
                return new DelimitedRecord(Type.JUNK, lineNo, line, null, -1, null, null, null);
            }

            current.append(line);
        }
    }

    private long getTimeStamp(final String[] fields) {
        if (timeStampPos < 0 || timeStampFormat == null) {
            return -1;
        }

        final DateFormat format = new SimpleDateFormat(timeStampFormat);
        final String timeStamp = fields[timeStampPos];
        Date parsed;

        try {
            parsed = format.parse(timeStamp);
        } catch (final java.text.ParseException e) {
            return -1;
        }

        return parsed.getTime();
    }

    private DelimitedRecord parse(final String line) {
        final String[] fields = getFields(line);

        if (maxPos >= fields.length) {
            return null;
        }

        final long timestamp = getTimeStamp(fields);

        if (timestamp < 0) {
            return null;
        }

        final String clazz = getField(fields, classPos);
        final String method = getField(fields, methodPos);
        final String level = getField(fields, levelPos);

        return new DelimitedRecord(Type.RECOGNIZED, lineNo, line, fields, timestamp, clazz, method, level);
    }
}
